/*
	BBitMapHandler.cpp
	
	Copyright 1996 Jeremy Moskovich, Digital Reality Studios
	All rights reserved.
	jeremy@vms.huji.ac.il
	
	A datatype for handling the Be ScreenDump file format, for use with John Watte's Datatype
	   library.
	   
	This DataType will only read in data into a DATABitmap structure, it doesn't know how to 
		create a view.
	
	It does not determine the contents of the inputted data by the type passed
	   into it, but instead by reading the first few bytes of a stream and seeing
	   if they are legal...
	   
    The format of the Be Screendump format is as follows:
    	Offset 10 - Width of image - 1
    	Offset 14 - Height of imag - 1
    	Offset 16 - color depth of image (currently one of B_RGB_32_BIT or B_COLOR_8_BIT)
    	Offset 20 - total size of bitmap data to follow
    	Offset 24 - bitMap data, with each row padded to fill longword boundries, ready
    	              to be pumped into a BBitmap file...
*/

#include "DataHandler.h"
#include "Datatypes.h"
#include "DataStreams.h"

//file type of Be Screen Dump
#define SCRNDMP_TYPE	'BITM'

//number of bytes to read at once
#define BUFSIZE		256

#define DEBUG 	0

#define SANITY( x )		if( (x) != B_NO_ERROR ) return FALSE

//============================= Function Protos =====================================
bool DoTranslation( DStream & inStream, DStream & outStream );

bool CanOpen( DStream & inStream );


char handlerName[] = "Be ScreenDump Datatype";
long handlerVersion = 100;
char handlerInfo[] = "Version 1.0, copyright Jeremy Moskovich (jeremy@vms.huji.ac.il) all rights reserved";



Format inputFormats[] = {
	{	SCRNDMP_TYPE ,	DATA_BITMAP,
		0.9,	1.0,
		"",
		"Be Screen Dump Data"
	},
	{	0,	0,
		0,	0,
		"",
		""
	}
};


/*
	This function is called by DataTypes to try and determine if we can translate to the
	specific format...
*/

long
Identify(
	DStream &			inSpec,
	const Format *		format,
	BMessage *			/* ioExtension */,
	DATAInfo &			outInfo,
	ulong				outType )
{
	#if DEBUG
		printf ("BBitMap Handler IDENTIFY Function called\n");
	#endif


//--------- If the caller passes in a hint, then if the hint isn't equal to the 
// Be Screendump creator, we don't try to open the file...
//  If no hint is passed in, then the file type isn't taken into consideration at all
//and the file's worthiness is determined by its content alone (via the canOpen function )
	
	if( format != 0 )
	{
		if( format->type != SCRNDMP_TYPE )
		{
			#if DEBUG
			     printf("format->type not correct\n");
			#endif
			
			return DATA_NO_HANDLER;
		}
		else
		{
			#if DEBUG
			     printf("format->type correct!!!\n");
		    #endif
		}
	}

	if( (outType == DATA_BITMAP) && (CanOpen( inSpec ) ) )
	{
        //I'm not sure if its correct to do things this way...
		outInfo.formatType = inputFormats[0].type;
		outInfo.formatGroup = inputFormats[0].t_cls;
		outInfo.formatQuality = inputFormats[0].quality;
		outInfo.formatCapability = inputFormats[0].capability;
		strcpy(outInfo.formatName, inputFormats[0].formatName);
		strcpy(outInfo.MIMEName, inputFormats[0].MIME);
		
		return 0;
	}
	else
	{
		return DATA_NO_HANDLER;
	}
}


bool 
CanOpen( DStream & inStream )
{
	ushort height, width;
	color_space bitDepth;
	ulong totalSize;

	SANITY( inStream.Seek( 10 ) );
	SANITY( inStream.Read( (void *)&width, 2 ) );
	
	SANITY( inStream.Seek( 14 ) );
	SANITY( inStream.Read( (void *)&height, 2 ) );

	SANITY( inStream.Seek( 16 ) );
	SANITY( inStream.Read( (void *)&bitDepth, 4 ) );

	SANITY( inStream.Seek( 20 ) );
	SANITY( inStream.Read( (void *)&totalSize, 4 ) );
	
	ulong rowBytes = width + 1;
	if( bitDepth == B_RGB_32_BIT )
		rowBytes *= 4;
	while( rowBytes & 3 ) rowBytes++;
	
	if( totalSize != ((height + 1) * rowBytes) )
	{
		#if DEBUG
			printf( "Total Size in file is incorrect\n" );
		#endif
		
		return FALSE;
	}
		
	if( ( width > 0 ) && (height > 0) && (bitDepth <= B_RGB_16_BIT) && (bitDepth > 0) )
	{//if theres nothing fishy about the file...
		
		#if DEBUG
		printf( "Image can be opened, dimensions are %d x %d x %d (deep) \n"
					, width, height, bitDepth );
		#endif
		
		return TRUE;
    }
    
    return FALSE;
}

bool
DoTranslation( DStream & inStream, DStream & outStream )
{
	ushort height, width;
	ulong rowBytes;
	
	color_space bitDepth;
	uchar *inBuf;
	ulong amtRead = BUFSIZE;

	ulong type = DATA_BITMAP;
	
	struct DATABitmap outSpec;
	
	#if DEBUG
		printf( "\tDoTranslate called OK\n" );
	#endif
	
	
	//Read in dimensions and butdepth from file...
	SANITY( inStream.Seek( 10 ) );
	SANITY( inStream.Read( &width, 2 ) );
	
	SANITY( inStream.Seek( 14 ) );
	inStream.Read( &height, 2 );
	
	SANITY( inStream.Seek( 16 ) );
	inStream.Read( &bitDepth, 4 );

	//calculate rowBytes
	rowBytes = width + 1;
	if( bitDepth == B_RGB_32_BIT )
		rowBytes *= 4;
	while( rowBytes & 3 ) rowBytes++;

    #if DEBUG
    	printf("Image has %d bytes per row\n", rowBytes);
    #endif

	//sanity check....
	if( ( width < 0 ) || (height < 0) || (bitDepth > B_RGB_16_BIT) || (bitDepth == 0) )
		return FALSE;
		
	#if DEBUG
		printf("\tDoTranslate: All initial Sanity checks passed OK\n");
	#endif
// F I L L  I N  DATA_BITMAP  S T R U C T U R E

	outSpec.magic = 'bits';
	outSpec.bounds = BRect( 0, 0, width, height );
	outSpec.rowBytes = rowBytes;
	outSpec.colors = bitDepth;
	outSpec.dataSize = ((height + 1) * rowBytes);
	
	SANITY( outStream.Seek( 0 ) );
	SANITY( outStream.Write( &outSpec, sizeof(struct DATABitmap) ) );

// S T A R T  R E A D I N G  I N  B I T M A P  D A T A
	//Go to start of bitMap data...		
	SANITY( inStream.Seek( 24 ) );

	inBuf = (uchar *) malloc( BUFSIZE );
	
	if( outStream.Seek( sizeof( DATABitmap ) ) != B_NO_ERROR )
	{
		#if DEBUG
		     printf("error trying to seek to past DataBitMap, size of struct is: %d (should be %d)\n",
		       outStream.Size(), sizeof( DATABitmap ) );
		#endif
		       
		outStream.SetSize(sizeof( DATABitmap ) + 1);
		outStream.Seek( sizeof( DATABitmap ) );
	}
	
	long err = 0;
	
	do 
	{	
		err = inStream.Read(inBuf, amtRead);
		
		if( outStream.Write( inBuf, amtRead ) != B_NO_ERROR )
		{//We couldn't write all output bytes
			#if DEBUG
				printf( "\tDoTranslation call to Write failed\n" );
			#endif
			
			free( inBuf );
			inBuf = 0;
			return FALSE;
		}
	} while( err == B_NO_ERROR );
	
	free( inBuf );
	inBuf = 0;
	
	return TRUE;
}

long
Translate(
	DStream &			inStream,
	const DATAInfo &	inInfo,
	BMessage *			/* ioExtension */,
	ulong				outFormat,
	DStream &			outStream)
{

	#if DEBUG
		printf ("BBitMap Handler TRANSLATE Function called\n");
    #endif
	
	//First make sure again that the file is Kosher
	if( ( (outFormat != DATA_BITMAP) && (outFormat != 0) ) || !CanOpen( inStream ) )
	{
		#if DEBUG
		     printf("\tTranslate() function failed initial sanity check\n");
        #endif
        
		return DATA_NO_HANDLER;
	}
	
	
	if( DoTranslation( inStream, outStream ) )
	{
		#if DEBUG
		     printf("\tTranslate was succesfull\n");
		#endif
		
		return B_NO_ERROR;
	}
	else
		return DATA_NO_HANDLER;
}